==> firebase_pagination-0.1.1-dev.tar.gz: lib/src/firestore_pagination.dart <==
// Flutter Packages
import 'package:flutter/material.dart';

// Dart Packages
import 'dart:async';

// Firebase Packages
import 'package:cloud_firestore/cloud_firestore.dart';

// Data Models
import 'models/view_type.dart';
import 'models/wrap_options.dart';

// Widgets
import 'widgets/defaults/bottom_loader.dart';
import 'widgets/defaults/empty_screen.dart';
import 'widgets/defaults/initial_loader.dart';
import 'widgets/views/build_pagination.dart';

/// A [StreamBuilder] that automatically loads more data when the user scrolls to the bottom.
///
/// Supports realtime updates and updates to loaded data.
class FirestorePagination extends StatefulWidget {
  /// Creates a [StreamBuilder] that automatically loads more data when the user scrolls to the bottom.
  ///
  /// Data can be represented in a [ListView], [GridView] or scollable [Wrap].
  const FirestorePagination({
    Key? key,
    required this.query,
    required this.itemBuilder,
    this.limit = 10,
    this.viewType = ViewType.list,
    this.isRealtime = true,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.wrapOptions = const WrapOptions(),
    this.onEmpty = const EmptyScreen(),
    this.bottomLoader = const BottomLoader(),
    this.initialLoader = const InitialLoader(),
  }) : super(key: key);

  /// The query to use to fetch data from Firestore.
  final Query query;

  /// The builder to use to build the items in the list.
  final Widget Function(BuildContext, DocumentSnapshot, int) itemBuilder;

  /// The number of items to fetch from Firestore at once.
  ///
  /// Defaults to `10`.
  final int limit;

  /// The type of view to use for the list.
  ///
  /// Defaults to [ViewType.list].
  final ViewType viewType;

  /// Whether to fetch newly added items as they are added to Firestore.
  ///
  /// Defaults to `true`.
  final bool isRealtime;

  /// The delegate to use for the [GridView].
  ///
  /// Defaults to [SliverGridDelegateWithFixedCrossAxisCount].
  final SliverGridDelegate gridDelegate;

  /// The [Wrap] widget properties to use.
  ///
  /// Defaults to [WrapOptions].
  final WrapOptions wrapOptions;

  /// The widget to use when data is empty.
  ///
  /// Defaults to [EmptyScreen].
  final Widget onEmpty;

  /// The widget to use when more data is loading.
  ///
  /// Defaults to [BottomLoader].
  final Widget bottomLoader;

  /// The widget to use when data is loading initially.
  ///
  /// Defaults to [InitialLoader].
  final Widget initialLoader;

  @override
  State<FirestorePagination> createState() => _FirestorePaginationState();
}

/// The state of the [FirestorePagination] widget.
class _FirestorePaginationState extends State<FirestorePagination> {
  /// All the data that has been loaded from Firestore.
  final List<DocumentSnapshot> _docs = [];

  /// Snapshot subscription for the query.
  ///
  /// Also handles updates to loaded data.
  StreamSubscription<QuerySnapshot>? _streamSub;

  /// Snapshot subscription for the query to handle newly added data.
  StreamSubscription<QuerySnapshot>? _realtimeStreamSub;

  /// [ScrollController] to listen to scroll end and load more data.
  final ScrollController _controller = ScrollController();

  /// Whether initial data is loading.
  bool _isInitialLoading = true;

  /// Whether more data is loading.
  bool _isFetching = false;

  /// Whether the end for given query has been reached.
  ///
  /// This is used to determine if more data should be loaded when the user scrolls to the bottom.
  bool _isEnded = false;

  /// Loads more data from Firestore and handles updates to loaded data.
  ///
  /// Setting [getMore] to `false` will only set listener for the currently loaded data.
  Future<void> _loadDocuments({bool getMore = true}) async {
    // To cancel previous updates listener when new one is set.
    final tempSub = _streamSub;

    if (getMore) setState(() => _isFetching = true);

    int docsLimit = _docs.length + (getMore ? widget.limit : 0);
    Query docsQuery = widget.query.limit(docsLimit);
    if (_docs.isNotEmpty) {
      docsQuery = docsQuery.startAtDocument(_docs.first);
    }

    _streamSub = docsQuery.snapshots().listen((QuerySnapshot snapshot) async {
      await tempSub?.cancel();

      _docs.clear();
      _docs.addAll(snapshot.docs);

      // To set new updates listener for the existing data
      // or to set new realtime listener if the first document is removed.
      bool isDocRemoved = snapshot.docChanges.any(
        (DocumentChange change) => change.type == DocumentChangeType.removed,
      );

      _isFetching = false;
      if (!_isEnded && !isDocRemoved) {
        _isEnded = snapshot.docs.length < docsLimit;
      }

      if (isDocRemoved || _isInitialLoading) {
        _isInitialLoading = false;
        if (snapshot.docs.isNotEmpty) {
          // Set updates listener for the existing data starting from the first document only.
          await _loadDocuments(getMore: false);
        } else {
          _streamSub?.cancel();
        }
        if (widget.isRealtime) _setRealtimeListner();
      }

      if (mounted) setState(() {});
    });
  }

  /// Sets the realtime listener for the query.
  ///
  /// Fires when new data is added to the query.
  Future<void> _setRealtimeListner() async {
    // To cancel previous realtime listener when new one is set.
    final tempSub = _realtimeStreamSub;

    Query latestDocQuery = widget.query.limit(1);
    if (_docs.isNotEmpty) {
      latestDocQuery = latestDocQuery.endBeforeDocument(_docs.first);
    }

    _realtimeStreamSub = latestDocQuery.snapshots().listen(
      (QuerySnapshot snapshot) async {
        await tempSub?.cancel();
        if (snapshot.docs.isEmpty ||
            snapshot.docs.first.metadata.hasPendingWrites) return;

        _docs.insert(0, snapshot.docs.first);

        // To handle newly added data after this curently loaded data.
        await _setRealtimeListner();

        // Set updates listener for the newly added data.
        _loadDocuments(getMore: false);
      },
    );
  }

  /// To handle scroll end event and load more data.
  void _scrollListener() {
    final ScrollPosition position = _controller.position;
    if (position.pixels >= (position.maxScrollExtent - 50) &&
        !_isInitialLoading &&
        !_isFetching &&
        !_isEnded) {
      _loadDocuments();
    }
  }

  @override
  void initState() {
    super.initState();
    _loadDocuments();
    _controller.addListener(_scrollListener);
  }

  @override
  void dispose() {
    _streamSub?.cancel();
    _realtimeStreamSub?.cancel();
    _controller.removeListener(_scrollListener);
    _controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _isInitialLoading
        ? widget.initialLoader
        : _docs.isEmpty
            ? widget.onEmpty
            : BuildPagination(
                controller: _controller,
                items: _docs,
                itemBuilder: widget.itemBuilder,
                isLoading: _isFetching,
                bottomLoader: widget.bottomLoader,
                gridDelegate: widget.gridDelegate,
                viewType: widget.viewType,
                wrapOptions: widget.wrapOptions,
              );
  }
}

==> firebase_pagination-0.2.0-dev.tar.gz: lib/src/firestore_pagination.dart <==
// Flutter Packages
import 'package:flutter/material.dart';

// Dart Packages
import 'dart:async';

// Firebase Packages
import 'package:cloud_firestore/cloud_firestore.dart';

// Data Models
import 'models/view_type.dart';
import 'models/wrap_options.dart';

// Widgets
import 'widgets/defaults/bottom_loader.dart';
import 'widgets/defaults/empty_screen.dart';
import 'widgets/defaults/initial_loader.dart';
import 'widgets/views/build_pagination.dart';

/// A [StreamBuilder] that automatically loads more data when the user scrolls
/// to the bottom.
///
/// Optimized for [FirebaseFirestore] with fields like `createdAt` and
/// `timestamp` to sort the data.
///
/// Supports realtime updates and updates to loaded data.
///
/// Data can be represented in a [ListView], [GridView] or scollable [Wrap].
class FirestorePagination extends StatefulWidget {
  /// Creates a [StreamBuilder] widget that automatically loads more data when
  /// the user scrolls to the bottom.
  ///
  /// Optimized for [FirebaseFirestore] with fields like `createdAt` and
  /// `timestamp` to sort the data.
  ///
  /// Supports realtime updates and updates to loaded data.
  ///
  /// Data can be represented in a [ListView], [GridView] or scollable [Wrap].
  const FirestorePagination({
    Key? key,
    required this.query,
    required this.itemBuilder,
    this.limit = 10,
    this.viewType = ViewType.list,
    this.isRealtime = true,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.wrapOptions = const WrapOptions(),
    this.onEmpty = const EmptyScreen(),
    this.bottomLoader = const BottomLoader(),
    this.initialLoader = const InitialLoader(),
  }) : super(key: key);

  /// The query to use to fetch data from Firestore.
  final Query query;

  /// The builder to use to build the items in the list.
  ///
  /// The builder is passed the build context, snapshot of the document and
  /// index of the item in the list.
  final Widget Function(BuildContext, DocumentSnapshot, int) itemBuilder;

  /// The number of items to fetch from Firestore at once.
  ///
  /// Defaults to `10`.
  final int limit;

  /// The type of view to use for the list.
  ///
  /// Defaults to [ViewType.list].
  final ViewType viewType;

  /// Whether to fetch newly added items as they are added to Firestore.
  ///
  /// Defaults to `true`.
  final bool isRealtime;

  /// The delegate to use for the [GridView].
  ///
  /// Defaults to [SliverGridDelegateWithFixedCrossAxisCount].
  final SliverGridDelegate gridDelegate;

  /// The [Wrap] widget properties to use.
  ///
  /// Defaults to [WrapOptions].
  final WrapOptions wrapOptions;

  /// The widget to use when data is empty.
  ///
  /// Defaults to [EmptyScreen].
  final Widget onEmpty;

  /// The widget to use when more data is loading.
  ///
  /// Defaults to [BottomLoader].
  final Widget bottomLoader;

  /// The widget to use when data is loading initially.
  ///
  /// Defaults to [InitialLoader].
  final Widget initialLoader;

  @override
  State<FirestorePagination> createState() => _FirestorePaginationState();
}

/// The state of the [FirestorePagination] widget.
class _FirestorePaginationState extends State<FirestorePagination> {
  /// All the data that has been loaded from Firestore.
  final List<DocumentSnapshot> _docs = [];

  /// Snapshot subscription for the query.
  ///
  /// Also handles updates to loaded data.
  StreamSubscription<QuerySnapshot>? _streamSub;

  /// Snapshot subscription for the query to handle newly added data.
  StreamSubscription<QuerySnapshot>? _realtimeStreamSub;

  /// [ScrollController] to listen to scroll end and load more data.
  final ScrollController _controller = ScrollController();

  /// Whether initial data is loading.
  bool _isInitialLoading = true;

  /// Whether more data is loading.
  bool _isFetching = false;

  /// Whether the end for given query has been reached.
  ///
  /// This is used to determine if more data should be loaded when the user
  /// scrolls to the bottom.
  bool _isEnded = false;

  /// Loads more data from Firestore and handles updates to loaded data.
  ///
  /// Setting [getMore] to `false` will only set listener for the currently
  /// loaded data.
  Future<void> _loadDocuments({bool getMore = true}) async {
    // To cancel previous updates listener when new one is set.
    final tempSub = _streamSub;

    if (getMore) setState(() => _isFetching = true);

    int docsLimit = _docs.length + (getMore ? widget.limit : 0);
    Query docsQuery = widget.query.limit(docsLimit);
    if (_docs.isNotEmpty) {
      docsQuery = docsQuery.startAtDocument(_docs.first);
    }

    _streamSub = docsQuery.snapshots().listen((QuerySnapshot snapshot) async {
      await tempSub?.cancel();

      _docs.clear();
      _docs.addAll(snapshot.docs);

      // To set new updates listener for the existing data
      // or to set new realtime listener if the first document is removed.
      bool isDocRemoved = snapshot.docChanges.any(
        (DocumentChange change) => change.type == DocumentChangeType.removed,
      );

      _isFetching = false;
      if (!_isEnded && !isDocRemoved) {
        _isEnded = snapshot.docs.length < docsLimit;
      }

      if (isDocRemoved || _isInitialLoading) {
        _isInitialLoading = false;
        if (snapshot.docs.isNotEmpty) {
          // Set updates listener for the existing data starting from the first
          // document only.
          await _loadDocuments(getMore: false);
        } else {
          _streamSub?.cancel();
        }
        if (widget.isRealtime) _setRealtimeListner();
      }

      if (mounted) setState(() {});
    });
  }

  /// Sets the realtime listener for the query.
  ///
  /// Fires when new data is added to the query.
  Future<void> _setRealtimeListner() async {
    // To cancel previous realtime listener when new one is set.
    final tempSub = _realtimeStreamSub;

    Query latestDocQuery = widget.query.limit(1);
    if (_docs.isNotEmpty) {
      latestDocQuery = latestDocQuery.endBeforeDocument(_docs.first);
    }

    _realtimeStreamSub = latestDocQuery.snapshots().listen(
      (QuerySnapshot snapshot) async {
        await tempSub?.cancel();
        if (snapshot.docs.isEmpty ||
            snapshot.docs.first.metadata.hasPendingWrites) return;

        _docs.insert(0, snapshot.docs.first);

        // To handle newly added data after this curently loaded data.
        await _setRealtimeListner();

        // Set updates listener for the newly added data.
        _loadDocuments(getMore: false);
      },
    );
  }

  /// To handle scroll end event and load more data.
  void _scrollListener() {
    final ScrollPosition position = _controller.position;
    if (position.pixels >= (position.maxScrollExtent - 50) &&
        !_isInitialLoading &&
        !_isFetching &&
        !_isEnded) {
      _loadDocuments();
    }
  }

  @override
  void initState() {
    super.initState();
    _loadDocuments();
    _controller.addListener(_scrollListener);
  }

  @override
  void dispose() {
    _streamSub?.cancel();
    _realtimeStreamSub?.cancel();
    _controller.removeListener(_scrollListener);
    _controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _isInitialLoading
        ? widget.initialLoader
        : _docs.isEmpty
            ? widget.onEmpty
            : BuildPagination(
                controller: _controller,
                items: _docs,
                itemBuilder: widget.itemBuilder,
                isLoading: _isFetching,
                bottomLoader: widget.bottomLoader,
                gridDelegate: widget.gridDelegate,
                viewType: widget.viewType,
                wrapOptions: widget.wrapOptions,
              );
  }
}
